# -*- coding: utf-8 -*-

"""Tests of waveform loader."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np
from numpy.testing import assert_array_equal as ae
from pytest import raises

from phylib.io.mock import artificial_traces, artificial_spike_samples
from ..waveform import _slice, WaveformLoader, WaveformExtractor


#------------------------------------------------------------------------------
# Tests extractor
#------------------------------------------------------------------------------

def test_extract_simple():
    weak = 1.
    strong = 2.
    nc = 4
    ns = 20

    data = np.random.uniform(size=(ns, nc), low=0., high=1.)

    data[10, 0] = 0.5
    data[11, 0] = 1.5
    data[12, 0] = 1.0

    data[10, 1] = 1.5
    data[11, 1] = 2.5
    data[12, 1] = 2.0

    component = np.array([[10, 0],
                          [10, 1],
                          [11, 0],
                          [11, 1],
                          [12, 0],
                          [12, 1],
                          ])

    we = WaveformExtractor(extract_before=3,
                           extract_after=5,
                           )
    we.set_thresholds(weak=weak, strong=strong)

    # _component()
    comp = we._component(component, n_samples=ns)
    ae(comp.comp_s, [10, 10, 11, 11, 12, 12])
    ae(comp.comp_ch, [0, 1, 0, 1, 0, 1])
    assert (comp.s_min, comp.s_max) == (10 - 3, 12 + 4)

    # _normalize()
    assert we._normalize(weak) == 0
    assert we._normalize(strong) == 1
    ae(we._normalize([(weak + strong) / 2.]), [.5])

    # _comp_wave()
    wave = we._comp_wave(data, comp)
    assert wave.shape == (3 + 5 + 1, nc)
    ae(wave[3:6, :], [[0.5, 1.5, 0., 0.],
                      [1.5, 2.5, 0., 0.],
                      [1.0, 2.0, 0., 0.]])

    # masks()
    masks = we.masks(data, wave, comp)
    ae(masks, [.5, 1., 0, 0])

    # spike_sample_aligned()
    s = we.spike_sample_aligned(wave, comp)
    assert 11 <= s < 12

    # extract()
    wave_e = we.extract(data, s)
    assert wave_e.shape[1] == wave.shape[1]
    ae(wave[3:6, :2], wave_e[3:6, :2])

    # align()
    wave_a = we.align(wave_e, s)
    assert wave_a.shape == (3 + 5, nc)

    # Test final call.
    s_f, masks_f, wave_f = we(component, data=data, data_t=data)
    assert s_f == s
    ae(masks_f, masks)
    ae(wave_f, wave_a)

    # Tests with a different order.
    we = WaveformExtractor(extract_before=3,
                           extract_after=5,
                           thresholds={'weak': weak,
                                       'strong': strong},
                           )
    s_f_o, masks_f_o, wave_f_o = we(component, data=data, data_t=data)
    assert s_f == s_f_o
    assert np.allclose(wave_f, wave_f_o)
    ae(masks_f_o, [0.5, 1., 0., 0.])


#------------------------------------------------------------------------------
# Tests utility functions
#------------------------------------------------------------------------------

def test_slice():
    assert _slice(0, (20, 20)) == slice(0, 20, None)


#------------------------------------------------------------------------------
# Tests loader
#------------------------------------------------------------------------------

def waveform_loader(do_filter=False, mask_threshold=None):
    n_samples_trace, n_channels = 1000, 5
    h = 10
    n_samples_waveforms = 2 * h
    n_spikes = n_samples_trace // (2 * n_samples_waveforms)
    sample_rate = 2000.

    traces = artificial_traces(n_samples_trace, n_channels)
    spike_samples = artificial_spike_samples(n_spikes,
                                             max_isi=2 * n_samples_waveforms)

    loader = WaveformLoader(traces=traces,
                            spike_samples=spike_samples,
                            n_samples_waveforms=n_samples_waveforms,
                            filter_order=3 if do_filter else None,
                            sample_rate=sample_rate,
                            )
    return loader


def test_loader_simple():
    loader = waveform_loader()
    spike_samples = loader.spike_samples
    traces = loader.traces
    n_samples_traces, n_channels = traces.shape
    n_samples_waveforms = loader.n_samples_waveforms
    h = n_samples_waveforms // 2

    waveforms = loader[10:20]
    assert waveforms.shape == (10, n_samples_waveforms, n_channels)
    t = spike_samples[15]
    w1 = waveforms[5, ...]
    w2 = traces[t - h:t + h, :]
    assert np.allclose(w1, w2)
    assert np.allclose(loader[15], w2)


def test_edges():
    loader = waveform_loader(do_filter=True)
    ns = loader.n_samples_waveforms + sum(loader._filter_margin)
    nc = loader.n_channels

    assert loader._load_at(0).shape == (ns, nc)
    assert loader._load_at(5).shape == (ns, nc)
    assert loader._load_at(loader.n_samples_trace - 5).shape == (ns, nc)
    assert loader._load_at(loader.n_samples_trace - 1).shape == (ns, nc)


def test_loader_filter_1():
    traces = np.c_[np.arange(20), np.arange(20, 40)].astype(np.int32)
    n_samples_trace, n_channels = traces.shape
    h = 3

    def my_filter(x, axis=0):
        return x * x

    loader = WaveformLoader(spike_samples=np.arange(20),
                            n_samples_waveforms=(h, h),
                            )
    assert loader[0].shape == (1, 2 * h, loader.n_channels)

    loader.traces = traces
    loader._filter = my_filter

    t = 10
    waveform_filtered = loader[t]
    traces_filtered = my_filter(traces)
    assert np.allclose(waveform_filtered, traces_filtered[t - h:t + h, :])


def test_loader_filter_2():
    loader = waveform_loader(do_filter=True)
    ns = loader.n_samples_waveforms
    nc = loader.n_channels

    with raises(ValueError):
        loader._load_at(-10)

    assert loader[0].shape == (1, ns, nc)
    assert loader[:].shape == (loader.n_spikes, ns, nc)


def test_loader_filter_3():
    loader = waveform_loader()
    ns = loader.n_samples_waveforms

    w = loader.get([0], channels=[0])
    assert w.shape == (1, ns, 1)



# -*- coding: utf-8 -*-

"""Test probe merging."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np

from ..merge import Merger
from phylib.io.alf import EphysAlfCreator
from phylib.io.array import _index_of
from phylib.io.model import load_model
from phylib.io.tests.conftest import _make_dataset


#------------------------------------------------------------------------------
# Merging tests
#------------------------------------------------------------------------------

def test_probe_merge_1(tempdir):
    out_dir = tempdir / 'merged'

    # Create two identical datasets.
    probe_names = ('probe_left', 'probe_right')
    for name in probe_names:
        (tempdir / name).mkdir(exist_ok=True, parents=True)
        _make_dataset(tempdir / name, param='dense', has_spike_attributes=False)

    subdirs = [tempdir / name for name in probe_names]

    # Merge them.
    m = Merger(subdirs, out_dir)
    single = load_model(tempdir / probe_names[0] / 'params.py')

    # Test the merged dataset.
    merged = m.merge()
    for name in ('n_spikes', 'n_channels', 'n_templates'):
        assert getattr(merged, name) == getattr(single, name) * 2
    assert merged.sample_rate == single.sample_rate


def test_probe_merge_2(tempdir):
    out_dir = tempdir / 'merged'

    # Create two identical datasets.
    probe_names = ('probe_left', 'probe_right')
    for name in probe_names:
        (tempdir / name).mkdir(exist_ok=True, parents=True)
        _make_dataset(tempdir / name, param='dense', has_spike_attributes=False)

    # Add small shift in the spike times of the second probe.
    single = load_model(tempdir / probe_names[0] / 'params.py')
    st_path = tempdir / 'probe_right/spike_times.npy'
    np.save(st_path, single.spike_samples + 1)

    subdirs = [tempdir / name for name in probe_names]

    # Merge them.
    m = Merger(subdirs, out_dir)
    merged = m.merge()

    # Test the merged dataset.
    for name in ('n_spikes', 'n_channels', 'n_templates'):
        assert getattr(merged, name) == getattr(single, name) * 2
    assert merged.sample_rate == single.sample_rate

    # Check the spikes.
    single = load_model(tempdir / probe_names[0] / 'params.py')
    assert np.allclose(single.spike_times, merged.spike_times[::2])
    assert np.allclose(merged.spike_times[1::2], single.spike_times + 4e-5)

    # Convert into ALF and load.
    alf = EphysAlfCreator(merged).convert(tempdir / 'alf')

    # Check that (almost) spike cluster probes are interleaved.
    cluster_ids = np.unique(alf.spike_clusters)
    sc_rel = _index_of(alf.spike_clusters, cluster_ids)
    sp = alf.cluster_probes[sc_rel]
    assert (sp[::2] != 0).sum() <= 1
    assert (sp[1::2] != 1).sum() <= 1



# -*- coding: utf-8 -*-

"""Waveform extraction."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import logging

import numpy as np
from scipy.interpolate import interp1d

from phylib.utils._types import _as_array, Bunch
from phylib.io.array import _pad, _get_padded, _range_from_slice
from phylib.traces.filter import apply_filter, bandpass_filter

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Waveform extractor from a connected component
#------------------------------------------------------------------------------

class WaveformExtractor(object):
    """Extract waveforms after data filtering and spike detection."""
    def __init__(
            self, extract_before=None, extract_after=None, weight_power=None, thresholds=None):
        self._extract_before = extract_before
        self._extract_after = extract_after
        self._weight_power = weight_power if weight_power is not None else 1.
        self._thresholds = thresholds or {}

    def _component(self, component, data=None, n_samples=None):
        comp_s = component[:, 0]  # shape: (component_size,)
        comp_ch = component[:, 1]  # shape: (component_size,)

        # Get the temporal window around the waveform.
        s_min, s_max = (comp_s.min() - 3), (comp_s.max() + 4)
        s_min = max(s_min, 0)
        s_max = min(s_max, n_samples)
        assert s_min < s_max

        return Bunch(comp_s=comp_s, comp_ch=comp_ch, s_min=s_min, s_max=s_max)

    def _normalize(self, x):
        x = _as_array(x)
        tw = self._thresholds['weak']
        ts = self._thresholds['strong']
        return np.clip((x - tw) / (ts - tw), 0, 1)

    def _comp_wave(self, data_t, comp):
        comp_s, comp_ch = comp.comp_s, comp.comp_ch
        s_min, s_max = comp.s_min, comp.s_max
        nc = data_t.shape[1]
        # Data on weak threshold crossings. shape: (some_length, nc)
        wave = np.zeros((s_max - s_min, nc), dtype=data_t.dtype)
        # The sample where the peak is reached, on each channel.
        wave[comp_s - s_min, comp_ch] = data_t[comp_s, comp_ch]
        return wave

    def masks(self, data_t, wave, comp):
        nc = data_t.shape[1]
        comp_ch = comp.comp_ch
        s_min = comp.s_min

        # Binary mask. shape: (nc,)
        masks_bin = np.zeros(nc, dtype=np.bool)
        masks_bin[np.unique(comp_ch)] = 1

        # Find the peaks (relative to the start of the chunk). shape: (nc,)
        peaks = np.argmax(wave, axis=0) + s_min
        # Peak values on each channel. shape: (nc,)
        peaks_values = data_t[peaks, np.arange(0, nc)] * masks_bin

        # Compute the float masks.
        masks_float = self._normalize(peaks_values)
        # Keep shank channels.
        return masks_float

    def spike_sample_aligned(self, wave, comp):
        s_min, s_max = comp.s_min, comp.s_max
        # Compute the fractional peak.
        wave_n = self._normalize(wave)
        wave_n_p = np.power(wave_n, self._weight_power)
        u = np.arange(s_max - s_min)[:, np.newaxis]
        # Spike aligned time relative to the start of the chunk.
        s_aligned = np.sum(wave_n_p * u) / np.sum(wave_n_p) + s_min
        return s_aligned

    def extract(self, data, s_aligned):
        s = int(s_aligned)
        # Get block of given size around peak sample.
        waveform = _get_padded(data, s - self._extract_before - 1, s + self._extract_after + 2)
        return waveform

    def align(self, waveform, s_aligned):
        s = int(s_aligned)
        sb, sa = self._extract_before, self._extract_after
        # Perform interpolation around the fractional peak.
        old_s = np.arange(s - sb - 1, s + sa + 2)
        new_s = np.arange(s - sb + 0, s + sa + 0) + (s_aligned - s)
        try:
            f = interp1d(old_s, waveform, bounds_error=True, kind='cubic', axis=0)
        except ValueError:  # pragma: no cover
            logger.warning("Interpolation error at time %d", s)
            return waveform
        return f(new_s)

    def set_thresholds(self, **kwargs):
        self._thresholds.update(kwargs)

    def __call__(self, component=None, data=None, data_t=None):
        assert data.shape == data_t.shape
        comp = self._component(component, data=data, n_samples=data_t.shape[0])

        wave = self._comp_wave(data_t, comp)
        masks = self.masks(data_t, wave, comp)
        s_aligned = self.spike_sample_aligned(wave, comp)

        waveform_unaligned = self.extract(data, s_aligned)
        waveform_aligned = self.align(waveform_unaligned, s_aligned)

        assert waveform_aligned.ndim == 2
        assert masks.ndim == 1
        assert waveform_aligned.shape[1] == masks.shape[0]

        return s_aligned, masks, waveform_aligned


#------------------------------------------------------------------------------
# Waveform loader from traces (used in the manual sorting GUI)
#------------------------------------------------------------------------------

def _before_after(n_samples):
    """Get the number of samples before and after."""
    if not isinstance(n_samples, (tuple, list)):
        before = n_samples // 2
        after = n_samples - before
    else:
        assert len(n_samples) == 2
        before, after = n_samples
        n_samples = before + after
    assert before >= 0
    assert after >= 0
    assert before + after == n_samples
    return before, after


def _slice(index, n_samples, margin=None):
    """Return a waveform slice."""
    if margin is None:
        margin = (0, 0)
    assert isinstance(n_samples, (tuple, list))
    assert len(n_samples) == 2
    before, after = n_samples
    assert isinstance(margin, (tuple, list))
    assert len(margin) == 2
    margin_before, margin_after = margin
    before += margin_before
    after += margin_after
    index = int(index)
    before = int(before)
    after = int(after)
    return slice(max(0, index - before), index + after, None)


class WaveformLoader(object):
    """Load waveforms from filtered or unfiltered traces."""

    def __init__(
            self, traces=None, sample_rate=None, spike_samples=None, filter_order=None,
            n_samples_waveforms=None):

        # Traces.
        if traces is not None:
            self.traces = traces
            self.n_samples_trace, self.n_channels = traces.shape
        else:
            self._traces = None
            self.n_samples_trace = self.n_channels = 0

        assert spike_samples is not None
        self._spike_samples = spike_samples
        self.n_spikes = len(spike_samples)

        # Define filter.
        if filter_order:
            filter_margin = filter_order * 3
            b_filter = bandpass_filter(
                rate=sample_rate, low=500., high=sample_rate * .475, order=filter_order)
            self._filter = lambda x, axis=0: apply_filter(x, b_filter, axis=axis)
        else:
            filter_margin = 0
            self._filter = lambda x, axis=0: x

        # Number of samples to return, can be an int or a
        # tuple (before, after).
        assert n_samples_waveforms is not None
        self.n_samples_before_after = _before_after(n_samples_waveforms)
        self.n_samples_waveforms = sum(self.n_samples_before_after)
        # Number of additional samples to use for filtering.
        self._filter_margin = _before_after(filter_margin)
        # Number of samples in the extracted raw data chunk.
        self._n_samples_extract = (self.n_samples_waveforms + sum(self._filter_margin))

        self.dtype = np.float32
        self.shape = (self.n_spikes, self._n_samples_extract, self.n_channels)
        self.ndim = 3

    @property
    def traces(self):
        """Raw traces."""
        return self._traces

    @traces.setter
    def traces(self, value):
        self.n_samples_trace, self.n_channels = value.shape
        self._traces = value

    @property
    def spike_samples(self):
        return self._spike_samples

    def _load_at(self, time, channels=None):
        """Load a waveform at a given time."""
        if channels is None:
            channels = slice(None, None, None)
        time = int(time)
        time_o = time
        ns = self.n_samples_trace
        if not (0 <= time_o < ns):
            raise ValueError("Invalid time {0:d}/{1:d}.".format(time_o, ns))
        slice_extract = _slice(time_o, self.n_samples_before_after, self._filter_margin)
        extract = self._traces[slice_extract][:, channels].astype(np.float32)

        # Pad the extracted chunk if needed.
        if slice_extract.start <= 0:
            extract = _pad(extract, self._n_samples_extract, 'left')
        elif slice_extract.stop >= ns - 1:
            extract = _pad(extract, self._n_samples_extract, 'right')

        assert extract.shape[0] == self._n_samples_extract
        return extract

    def get(self, spike_ids, channels=None):
        """Load the waveforms of the specified spikes."""
        if isinstance(spike_ids, slice):
            spike_ids = _range_from_slice(spike_ids, start=0, stop=self.n_spikes)
        if not hasattr(spike_ids, '__len__'):
            spike_ids = [spike_ids]
        if channels is None:
            channels = slice(None, None, None)
            nc = self.n_channels
        else:
            channels = np.asarray(channels, dtype=np.int32)
            assert np.all(channels < self.n_channels)
            nc = len(channels)

        # Ensure a list of time samples are being requested.
        spike_ids = _as_array(spike_ids)
        n_spikes = len(spike_ids)

        # Initialize the array.
        # NOTE: last dimension is time to simplify things.
        shape = (n_spikes, nc, self._n_samples_extract)
        waveforms = np.zeros(shape, dtype=np.float32)

        # No traces: return null arrays.
        if self.n_samples_trace == 0:
            return np.transpose(waveforms, (0, 2, 1))

        # Load all spikes.
        for i, spike_id in enumerate(spike_ids):
            assert 0 <= spike_id < self.n_spikes
            time = self._spike_samples[spike_id]

            # Extract the waveforms on the unmasked channels.
            try:
                w = self._load_at(time, channels)
            except ValueError as e:  # pragma: no cover
                logger.warning("Error while loading waveform: %s", str(e))
                continue

            assert w.shape == (self._n_samples_extract, nc)

            waveforms[i, :, :] = w.T

        # Filter the waveforms.
        waveforms_f = waveforms.reshape((-1, self._n_samples_extract))
        # Only filter the non-zero waveforms.
        unmasked = waveforms_f.max(axis=1) != 0
        waveforms_f[unmasked] = self._filter(waveforms_f[unmasked], axis=1)
        waveforms_f = waveforms_f.reshape((n_spikes, nc, self._n_samples_extract))

        # Remove the margin.
        margin_before, margin_after = self._filter_margin
        if margin_after > 0:
            assert margin_before >= 0
            waveforms_f = waveforms_f[:, :, margin_before:-margin_after]

        assert waveforms_f.shape == (n_spikes, nc, self.n_samples_waveforms)

        # NOTE: we transpose before returning the array.
        return np.transpose(waveforms_f, (0, 2, 1))

    def __getitem__(self, spike_ids):
        return self.get(spike_ids)



# -*- coding: utf-8 -*-

"""ALF dataset generation."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import logging
from pathlib import Path
import shutil
import ast

from tqdm import tqdm
import numpy as np

from phylib.utils._misc import _read_tsv_simple, ensure_dir_exists
from phylib.io.array import _spikes_per_cluster, select_spikes, _unique, grouped_mean, _index_of
from phylib.io.model import load_model

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# File utils
#------------------------------------------------------------------------------


# ## TODO

# probes.insertion
# probes.sitePositions
# probes.rawFilename
# channels.brainLocation

_FILE_RENAMES = [  # file_in, file_out, squeeze (bool to squeeze vector from matlab in npy)
    ('params.py', 'params.py', None),
    ('spike_clusters.npy', 'spikes.clusters.npy', True),
    ('amplitudes.npy', 'spikes.amps.npy', True),
    ('channel_positions.npy', 'channels.sitePositions.npy', False),
    ('templates.npy', 'clusters.templateWaveforms.npy', False),
    ('channel_map.npy', 'channels.rawRow.npy', True),
    ('channel_map.npy', 'channels.rawRow.npy', True),
    ('channel_probe.npy', 'channels.probes.npy', True),
    ('cluster_probes.npy', 'clusters.probes.npy', True),
    ('cluster_shanks.npy', 'clusters.shanks.npy', True),

    # ('probes.description.txt', 'probes.description.txt', False),
    # ('spike_templates.npy', 'ks2/spikes.clusters.npy', True),
    # ('cluster_ContamPct.tsv', 'ks2/clusters.ContamPct.tsv', False),
    # ('cluster_group.tsv', 'ks2/clusters.phyAnnotation.tsv', False),
    # ('cluster_KSLabel.tsv', 'ks2/clusters.group.tsv', False),
]

FILE_DELETES = [
    'temp_wh.dat',  # potentially large file that will clog the servers
]


def _read_npy_header(filename):
    d = {}
    with open(filename, 'rb') as fid:
        d['magic_string'] = fid.read(6)
        d['version'] = fid.read(2)
        d['len'] = int.from_bytes(fid.read(2), byteorder='little')
        d = {**d, **ast.literal_eval(fid.read(d['len']).decode())}
    return d


def _create_if_possible(path, new_path, force=False):
    """Prepare the copy/move/symlink of a file, by making sure the source exists
    while the destination does not."""
    if not Path(path).exists():  # pragma: no cover
        logger.warning("Path %s does not exist, skipping.", path)
        return False
    if Path(new_path).exists() and not force:  # pragma: no cover
        logger.warning("Path %s already exists, skipping.", new_path)
        return False
    ensure_dir_exists(new_path.parent)
    return True


def _copy_if_possible(path, new_path, force=False):
    if not _create_if_possible(path, new_path, force=force):
        return False
    logger.debug("Copying %s to %s.", path, new_path)
    shutil.copy(path, new_path)
    return True


def _load(path):
    path = str(path)
    if path.endswith('.npy'):
        return np.load(path)
    elif path.endswith(('.csv', '.tsv')):
        return _read_tsv_simple(path)[1]  # the function returns a tuple (field, data)
    elif path.endswith('.bin'):
        # TODO: configurable dtype
        return np.fromfile(path, np.int16)


#------------------------------------------------------------------------------
# Ephys ALF creator
#------------------------------------------------------------------------------

class EphysAlfCreator(object):
    """Class for converting a dataset in KS/phy format into ALF."""

    def __init__(self, model):
        self.model = model
        self.dir_path = Path(model.dir_path)
        self.spc = _spikes_per_cluster(model.spike_clusters)

    def convert(self, out_path, force=False):
        """Convert from KS/phy format to ALF."""
        logger.info("Converting dataset to ALF.")
        self.out_path = Path(out_path)
        if self.out_path.resolve() == self.dir_path.resolve():
            raise IOError("The source and target directories cannot be the same.")
        if not self.out_path.exists():
            self.out_path.mkdir()

        with tqdm(desc="Converting to ALF", total=60) as bar:
            self.copy_files(force=force)
            bar.update(10)
            self.make_spike_times()
            bar.update(10)
            self.make_cluster_waveforms()
            bar.update(10)
            self.make_depths()
            bar.update(10)
            self.make_mean_waveforms()
            bar.update(10)
            self.rm_files()
            bar.update(10)

        # Return the TemplateModel of the converted ALF dataset if the params.py file exists.
        params_path = self.out_path / 'params.py'
        if params_path.exists():
            return load_model(params_path)

    def copy_files(self, force=False):
        for fn0, fn1, squeeze in _FILE_RENAMES:
            f0 = self.dir_path / fn0
            f1 = self.out_path / fn1
            _copy_if_possible(f0, f1, force=force)
            if f0.exists() and squeeze and f0.suffix == '.npy':
                h = _read_npy_header(f0)
                # ks2 outputs vectors as multidimensional arrays. If there is no distinction
                # for Matlab, there is one in Numpy
                if len(h['shape']) == 2 and h['shape'][-1] == 1:
                    d = np.load(f0)
                    np.save(f1, d.squeeze())
                    continue

    def rm_files(self):
        for fn0 in FILE_DELETES:
            fn = self.dir_path.joinpath(fn0)
            if fn.exists():  # pragma: no cover
                fn.unlink()

    # File creation
    # -------------------------------------------------------------------------

    def _save_npy(self, filename, arr):
        """Save an array into a .npy file."""
        np.save(self.out_path / filename, arr)

    def make_spike_times(self):
        """We cannot just rename/copy spike_times.npy because it is in unit of
        *samples*, and not in seconds."""
        self.dir_path
        self._save_npy('spikes.times.npy', self.model.spike_times)

    def make_cluster_waveforms(self):
        """Return the channel index with the highest template amplitude, for
        every template."""
        p = self.dir_path
        tmp = self.model.sparse_templates.data

        peak_channel_path = p / 'clusters.peakChannel.npy'
        if not peak_channel_path.exists():
            # Create the cluster channels file.
            n_templates, n_samples, n_channels = tmp.shape

            # Compute the peak channels for each template.
            template_peak_channels = np.argmax(tmp.max(axis=1) - tmp.min(axis=1), axis=1)
            assert template_peak_channels.shape == (n_templates,)
            self._save_npy(peak_channel_path.name, template_peak_channels)
        else:  # pragma: no cover
            template_peak_channels = np.load(peak_channel_path)

        waveform_duration_path = p / 'clusters.waveformDuration.npy'
        if not waveform_duration_path.exists():
            # Compute the peak channel waveform for each template.
            waveforms = tmp[:, :, template_peak_channels]
            durations = waveforms.argmax(axis=1) - waveforms.argmin(axis=1)
            self._save_npy(waveform_duration_path.name, durations)

    def make_depths(self):
        """Make spikes.depths.npy, clusters.depths.npy."""
        channel_positions = self.model.channel_positions
        assert channel_positions.ndim == 2

        spike_clusters = self.model.spike_clusters
        assert spike_clusters.ndim == 1
        n_spikes = spike_clusters.shape[0]
        self.cluster_ids = _unique(self.model.spike_clusters)

        cluster_channels = np.load(self.out_path / 'clusters.peakChannel.npy')
        assert cluster_channels.ndim == 1
        n_clusters = cluster_channels.shape[0]

        clusters_depths = channel_positions[cluster_channels, 1]
        assert clusters_depths.shape == (n_clusters,)

        spike_clusters_rel = _index_of(spike_clusters, self.cluster_ids)
        assert spike_clusters_rel.max() < clusters_depths.shape[0]
        spikes_depths = clusters_depths[spike_clusters_rel]
        assert spikes_depths.shape == (n_spikes,)

        self._save_npy('spikes.depths.npy', spikes_depths)
        self._save_npy('clusters.depths.npy', clusters_depths)

    def make_mean_waveforms(self):
        """Make the mean waveforms file."""
        spike_ids = select_spikes(
            cluster_ids=self.cluster_ids,
            max_n_spikes_per_cluster=100,
            spikes_per_cluster=lambda clu: self.spc[clu],
            subset='random')
        waveforms = self.model.get_waveforms(spike_ids, np.arange(self.model.n_channels))
        try:
            mean_waveforms = grouped_mean(waveforms, self.model.spike_clusters[spike_ids])
            self._save_npy('clusters.meanWaveforms.npy', mean_waveforms)
        except IndexError as e:  # pragma: no cover
            logger.warning("Failed to create the mean waveforms file: %s.", e)



# -*- coding: utf-8 -*-

"""Template model."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import logging
import os.path as op
from operator import itemgetter
from pathlib import Path
import shutil

import numpy as np
import scipy.io as sio

from .array import _concatenate_virtual_arrays, _index_of, _spikes_in_clusters
from phylib.traces import WaveformLoader
from phylib.utils import Bunch
from phylib.utils._misc import _write_tsv_simple, _read_tsv_simple, read_python
from phylib.utils.geometry import linear_positions

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Utility functions
#------------------------------------------------------------------------------

def read_array(path, mmap_mode=None):
    """Read a binary array in npy or mat format, avoiding nan and inf values."""
    path = Path(path)
    arr_name = path.name
    ext = path.suffix
    if ext == '.mat':  # pragma: no cover
        out = sio.loadmat(path)[arr_name]
    elif ext == '.npy':
        out = np.load(path, mmap_mode=mmap_mode)
    # Filter out nan and inf values.
    # NOTE: do not check for nan/inf values on mmap arrays.
    # TODO: virtual memmap array where the replacement is done on-the-fly when reading the array.
    if mmap_mode is None:
        for w in ('nan', 'inf'):
            errors = getattr(np, 'is' + w)(out)
            if np.any(errors):
                n = np.sum(errors)
                n_tot = errors.size
                logger.warning("%d/%d values are %s in %s, replacing by zero.", n, n_tot, w, path)
                out[errors] = 0
    return out


def write_array(name, arr):
    """Save an array to a binary file."""
    np.save(name, arr)


def from_sparse(data, cols, channel_ids):
    """Convert a sparse structure into a dense one.

    Parameters
    ----------

    data : array-like
        A (n_spikes, n_channels_loc, ...) array with the data.
    cols : array-like
        A (n_spikes, n_channels_loc) array with the channel indices of
        every row in data.
    channel_ids : array-like
        List of requested channel ids (columns).

    """
    # The axis in the data that contains the channels.
    if len(channel_ids) != len(np.unique(channel_ids)):
        raise NotImplementedError("Multiple identical requested channels "
                                  "in from_sparse().")
    channel_axis = 1
    shape = list(data.shape)
    assert data.ndim >= 2
    assert cols.ndim == 2
    assert data.shape[:2] == cols.shape
    n_spikes, n_channels_loc = shape[:2]
    # NOTE: we ensure here that `col` contains integers.
    c = cols.flatten().astype(np.int32)
    # Remove columns that do not belong to the specified channels.
    c[~np.in1d(c, channel_ids)] = -1
    assert np.all(np.in1d(c, np.r_[channel_ids, -1]))
    # Convert column indices to relative indices given the specified
    # channel_ids.
    cols_loc = _index_of(c, np.r_[channel_ids, -1]).reshape(cols.shape)
    assert cols_loc.shape == (n_spikes, n_channels_loc)
    n_channels = len(channel_ids)
    # Shape of the output array.
    out_shape = shape
    # The channel dimension contains the number of requested channels.
    # The last column contains irrelevant values.
    out_shape[channel_axis] = n_channels + 1
    out = np.zeros(out_shape, dtype=data.dtype)
    x = np.tile(np.arange(n_spikes)[:, np.newaxis],
                (1, n_channels_loc))
    assert x.shape == cols_loc.shape == data.shape[:2]
    out[x, cols_loc, ...] = data
    # Remove the last column with values outside the specified
    # channels.
    out = out[:, :-1, ...]
    return out


def load_metadata(filename):
    """Load cluster metadata from a CSV file.

    Return (field_name, dictionary).

    """
    return _read_tsv_simple(filename)


def save_metadata(filename, field_name, metadata):
    """Save metadata in a CSV file."""
    return _write_tsv_simple(filename, field_name, metadata)


#------------------------------------------------------------------------------
# Raw data functions
#------------------------------------------------------------------------------

def _dat_n_samples(filename, dtype=None, n_channels=None, offset=None):
    """Get the number of samples from the size of a dat file."""
    assert dtype is not None
    item_size = np.dtype(dtype).itemsize
    offset = offset if offset else 0
    n_samples = (op.getsize(str(filename)) - offset) // (item_size * n_channels)
    assert n_samples >= 0
    return n_samples


def _dat_to_traces(dat_path, n_channels=None, dtype=None, offset=None):
    """Memmap a dat file."""
    assert dtype is not None
    assert n_channels is not None
    n_samples = _dat_n_samples(dat_path, n_channels=n_channels, dtype=dtype, offset=offset)
    return np.memmap(str(dat_path), dtype=dtype, shape=(n_samples, n_channels), offset=offset)


def load_raw_data(path=None, n_channels_dat=None, dtype=None, offset=None):
    """Load raw data at a given path."""
    if not path:
        return
    path = Path(path)
    if not path.exists():
        logger.warning("Path %s does not exist, trying ephys.raw filename.", path)
        path = path.parent / ('ephys.raw' + path.suffix)
        if not path.exists():
            logger.warning("Error while loading data: File `%s` not found.", path)
            return None
    assert path.exists()
    logger.debug("Loading traces at `%s`.", path)
    dtype = dtype if dtype is not None else np.int16
    return _dat_to_traces(path, n_channels=n_channels_dat, dtype=dtype, offset=offset)


#------------------------------------------------------------------------------
# Channel util functions
#------------------------------------------------------------------------------

def _all_positions_distinct(positions):
    """Return whether all positions are distinct."""
    return len(set(tuple(row) for row in positions)) == len(positions)


def get_closest_channels(channel_positions, channel_index, n=None):
    """Get the channels closest to a given channel on the probe."""
    x = channel_positions[:, 0]
    y = channel_positions[:, 1]
    x0, y0 = channel_positions[channel_index]
    d = (x - x0) ** 2 + (y - y0) ** 2
    out = np.argsort(d)
    if n:
        out = out[:n]
    assert out[0] == channel_index
    return out


#------------------------------------------------------------------------------
# I/O util functions
#------------------------------------------------------------------------------

def _find_first_existing_path(*paths, multiple_ok=True):
    out = []
    for path in paths:
        path = Path(path)
        if path.exists():
            out.append(path)
    if len(out) >= 2 and not multiple_ok:  # pragma: no cover
        raise IOError("Multiple conflicting files exist: %s." % ', '.join((out, path)))
    elif len(out) >= 1:
        return out[0]
    else:
        return None


def _close_memmap(name, obj):
    """Close a memmap array or a list of memmap arrays."""
    if isinstance(obj, np.memmap):
        logger.debug("Close memmap array %s.", name)
        obj._mmap.close()
    elif getattr(obj, 'arrs', None) is not None:
        # Support ConcatenatedArrays.
        _close_memmap('%s.arrs' % name, obj.arrs)
    elif isinstance(obj, (list, tuple)):
        [_close_memmap('%s[]' % name, item) for item in obj]
    elif isinstance(obj, dict):
        [_close_memmap('%s.%s' % (name, n), item) for n, item in obj.items()]


#------------------------------------------------------------------------------
# Template model
#------------------------------------------------------------------------------

class TemplateModel(object):
    """Object holding all data of a KiloSort/phy dataset.

    Constructor
    -----------

    dir_path : str or Path
        Path to the dataset directory
    dat_path : str, Path, or list
        Path to the raw data files.
    dtype : NumPy dtype
        Data type of the raw data file
    offset : int
        Header offset of the binary file
    n_channels_dat : int
        Number of channels in the dat file
    sample_rate : float
        Sampling rate of the data file.
    filter_order : int
        Order of the filter used for waveforms
    hp_filtered : bool
        Whether the raw data file is already high-pass filtered. In that case, disable the
        filtering for the waveform extraction.

    """

    """Number of closest channels used for templates."""
    n_closest_channels = 12

    """Fraction of the peak amplitude required by the closest channels to be kept as best
    channels."""
    amplitude_threshold = 0

    def __init__(self, **kwargs):
        # Default empty values.
        self.dat_path = []
        self.sample_rate = None
        self.n_channels_dat = None

        self.__dict__.update(kwargs)

        # Set dir_path.
        self.dir_path = Path(self.dir_path).resolve()
        assert isinstance(self.dir_path, Path)
        assert self.dir_path.exists()

        # Set dat_path.
        if not self.dat_path:  # pragma: no cover
            self.dat_path = []
        elif not isinstance(self.dat_path, (list, tuple)):
            self.dat_path = [self.dat_path]
        assert isinstance(self.dat_path, (list, tuple))
        self.dat_path = [Path(_).resolve() for _ in self.dat_path]

        self.dtype = getattr(self, 'dtype', np.int16)
        if not self.sample_rate:  # pragma: no cover
            logger.warning("No sample rate was given! Defaulting to 1 Hz.")
        self.sample_rate = float(self.sample_rate or 1.)
        assert self.sample_rate > 0
        self.offset = getattr(self, 'offset', 0)

        self.filter_order = None if getattr(self, 'hp_filtered', False) else 3

        self._load_data()
        self.waveform_loader = self._create_waveform_loader()

    #--------------------------------------------------------------------------
    # Internal loading methods
    #--------------------------------------------------------------------------

    def _load_data(self):
        """Load all data."""
        sr = self.sample_rate

        # Spikes.
        self.spike_samples = self._load_spike_samples()
        self.spike_times = self.spike_samples / sr
        ns, = self.n_spikes, = self.spike_times.shape

        # Spike amplitudes.
        self.amplitudes = self._load_amplitudes()
        if self.amplitudes is not None:
            assert self.amplitudes.shape == (ns,)

        # Spike templates.
        self.spike_templates = self._load_spike_templates()
        assert self.spike_templates.shape == (ns,)

        # Spike clusters.
        self.spike_clusters = self._load_spike_clusters()
        assert self.spike_clusters.shape == (ns,)

        # Channels.
        self.channel_mapping = self._load_channel_map()
        self.n_channels = nc = self.channel_mapping.shape[0]
        if self.n_channels_dat:
            assert np.all(self.channel_mapping <= self.n_channels_dat - 1)

        # Channel positions.
        self.channel_positions = self._load_channel_positions()
        assert self.channel_positions.shape == (nc, 2)
        if not _all_positions_distinct(self.channel_positions):  # pragma: no cover
            logger.error(
                "Some channels are on the same position, please check the channel positions file.")
            self.channel_positions = linear_positions(nc)

        # Channel shanks.
        self.channel_shanks = self._load_channel_shanks()
        assert self.channel_shanks.shape == (nc,)

        # Channel probes.
        self.channel_probes = self._load_channel_probes()
        assert self.channel_probes.shape == (nc,)
        self.probes = np.unique(self.channel_probes)
        self.n_probes = len(self.probes)

        # Ordering of the channels in the trace view.
        self.channel_vertical_order = np.argsort(self.channel_positions[:, 1], kind='mergesort')

        # Templates.
        self.sparse_templates = self._load_templates()
        self.n_templates, self.n_samples_waveforms, self.n_channels_loc = \
            self.sparse_templates.data.shape
        if self.sparse_templates.cols is not None:
            assert self.sparse_templates.cols.shape == (self.n_templates, self.n_channels_loc)

        # Cluster probes and shanks (only when loading an ALF dataset).
        self.cluster_probes = self._load_cluster_probes()
        self.cluster_shanks = self._load_cluster_shanks()

        # Whitening.
        try:
            self.wm = self._load_wm()
        except IOError:
            logger.debug("Whitening matrix file not found.")
            self.wm = np.eye(nc)
        assert self.wm.shape == (nc, nc)
        try:
            self.wmi = self._load_wmi()
        except IOError:
            logger.debug("Whitening matrix inverse file not found, computing it.")
            self.wmi = self._compute_wmi(self.wm)
        assert self.wmi.shape == (nc, nc)

        # Similar templates.
        self.similar_templates = self._load_similar_templates()
        assert self.similar_templates.shape == (self.n_templates, self.n_templates)

        # Traces and duration.
        self.traces = self._load_traces(self.channel_mapping)
        if self.traces is not None:
            self.duration = self.traces.shape[0] / float(self.sample_rate)
        else:
            self.duration = self.spike_times[-1]
        if self.spike_times[-1] > self.duration:  # pragma: no cover
            logger.debug(
                "There are %d/%d spikes after the end of the recording.",
                np.sum(self.spike_times > self.duration), self.n_spikes)

        # Features.
        self.sparse_features = self._load_features()
        self.features = self.sparse_features.data if self.sparse_features else None
        if self.sparse_features is not None:
            self.n_features_per_channel = self.sparse_features.data.shape[2]

        # Template features.
        self.sparse_template_features = self._load_template_features()
        self.template_features = (
            self.sparse_template_features.data if self.sparse_template_features else None)

        # Spike attributes.
        self.spike_attributes = self._load_spike_attributes()

        # Metadata.
        self.metadata = self._load_metadata()

    def _create_waveform_loader(self):
        # Number of time samples in the templates.
        nsw = self.n_samples_waveforms
        if self.traces is not None:
            return WaveformLoader(
                traces=self.traces,
                spike_samples=self.spike_samples,
                n_samples_waveforms=nsw,
                filter_order=self.filter_order,
                sample_rate=self.sample_rate,
            )

    def _find_path(self, *names, multiple_ok=True):
        return _find_first_existing_path(
            *(self.dir_path / name for name in names), multiple_ok=multiple_ok)

    def _read_array(self, path, mmap_mode=None):
        if not path:
            raise IOError(path)
        return read_array(path, mmap_mode=mmap_mode).squeeze()

    def _write_array(self, path, arr):
        return write_array(path, arr)

    def _load_metadata(self):
        """Load cluster metadata from all CSV/TSV files in the data directory."""
        # Files to exclude.
        excluded_names = ('cluster_info',)
        # Get all CSV/TSV files in the directory.
        files = list(self.dir_path.glob('*.csv'))
        files.extend(self.dir_path.glob('*.tsv'))
        metadata = {}
        for filename in files:
            if filename.stem in excluded_names:
                continue
            logger.debug("Load `%s`.", filename)
            try:
                field_name, values = load_metadata(filename)
            except Exception as e:
                logger.debug("Could not load %s: %s.", filename, str(e))
                continue
            metadata[field_name] = values
        return metadata

    #--------------------------------------------------------------------------
    # Specific loading methods
    #--------------------------------------------------------------------------

    def _load_spike_attributes(self):
        """Load all spike_*.npy files, called spike attributes."""
        files = list(self.dir_path.glob('spike_*.npy'))
        spike_attributes = Bunch()
        for filename in files:
            # The part after spike_***
            n = filename.stem[6:]
            # Skip known files.
            if n in ('clusters', 'templates', 'samples', 'times', 'amplitudes'):
                continue
            try:
                arr = self._read_array(filename)
                assert arr.shape[0] == self.n_spikes
                logger.debug("Load %s.", filename)
            except (IOError, AssertionError) as e:
                logger.warning("Unable to open %s: %s.", filename, e)
                continue
            spike_attributes[n] = arr
        return spike_attributes

    def _load_channel_map(self):
        path = self._find_path('channel_map.npy', 'channels.rawRow.npy')
        out = self._read_array(path)
        out = np.atleast_1d(out)
        assert out.ndim == 1
        assert out.dtype in (np.uint32, np.int32, np.int64)
        return out

    def _load_channel_positions(self):
        path = self._find_path('channel_positions.npy', 'channels.sitePositions.npy')
        out = self._read_array(path)
        out = np.atleast_2d(out)
        assert out.ndim == 2
        return out

    def _load_channel_probes(self):
        try:
            path = self._find_path('channel_probe.npy', 'channels.probes.npy')
            out = self._read_array(path)
            out = np.atleast_1d(out)
            assert out.ndim == 1
            return out
        except IOError:
            return np.zeros(self.n_channels, dtype=np.int32)

    def _load_cluster_probes(self):
        try:
            path = self._find_path('cluster_probes.npy', 'clusters.probes.npy')
            out = self._read_array(path)
            out = np.atleast_1d(out)
            assert out.ndim == 1
            return out
        except IOError:
            return np.zeros(self.n_templates, dtype=np.int32)

    def _load_cluster_shanks(self):  # pragma: no cover
        try:
            path = self._find_path('cluster_shanks.npy', 'clusters.shanks.npy')
            out = self._read_array(path)
            out = np.atleast_1d(out)
            assert out.ndim == 1
            return out
        except IOError:
            return np.zeros(self.n_templates, dtype=np.int32)

    def _load_channel_shanks(self):
        try:
            path = self._find_path('channel_shanks.npy', 'channels.shanks.npy')
            out = self._read_array(path).reshape((-1,))
            assert out.ndim == 1
            return out
        except IOError:
            logger.debug("No channel shank file found.")
            return np.zeros(self.n_channels, dtype=np.int32)

    def _load_traces(self, channel_map=None):
        if not self.dat_path:
            return
        paths = self.dat_path
        # Make sure we have a list of paths (virtually-concatenated).
        assert isinstance(paths, (list, tuple))
        n = self.n_channels_dat
        # Memmap all dat files and concatenate them virtually.
        traces = [
            load_raw_data(path, n_channels_dat=n, dtype=self.dtype, offset=self.offset)
            for path in paths]
        traces = [_ for _ in traces if _ is not None]
        traces = _concatenate_virtual_arrays(traces, channel_map)
        return traces

    def _load_amplitudes(self):
        try:
            out = self._read_array(self._find_path('amplitudes.npy', 'spikes.amps.npy'))
            assert out.ndim == 1
            return out
        except IOError:
            logger.debug("No amplitude file found.")
            return

    def _load_spike_templates(self):
        path = self._find_path('spike_templates.npy', 'spikes.clusters.npy')
        out = self._read_array(path)
        if out.dtype in (np.float32, np.float64):  # pragma: no cover
            out = out.astype(np.int32)
        assert out.dtype in (np.uint32, np.int32, np.int64)
        assert out.ndim == 1
        return out

    def _load_spike_clusters(self):
        path = self._find_path('spike_clusters.npy', 'spikes.clusters.npy', multiple_ok=False)
        if path is None:
            # Create spike_clusters file if it doesn't exist.
            tmp_path = self._find_path('spike_templates.npy', 'spikes.clusters.npy')
            path = self.dir_path / 'spike_clusters.npy'
            logger.debug("Copying from %s to %s.", tmp_path, path)
            shutil.copy(tmp_path, path)
        assert path.exists()
        logger.debug("Loading spike clusters.")
        # NOTE: we make a copy in memory so that we can update this array
        # during manual clustering.
        out = self._read_array(path).astype(np.int32)
        assert out.ndim == 1
        return out

    def _load_spike_samples(self):
        # WARNING: "spike_times.npy" is in units of samples. Need to
        # divide by the sampling rate to get spike times in seconds.
        path = self.dir_path / 'spike_times.npy'
        if path.exists():
            out = self._read_array(path)
        else:
            # WARNING: spikes.times.npy is in seconds, not samples !
            path = self.dir_path / 'spikes.times.npy'
            logger.info("Loading spikes.times.npy in seconds, converting to samples.")
            spike_times = self._read_array(path)
            out = (spike_times * self.sample_rate).astype(np.uint64)
        assert out.ndim == 1
        return out

    def _load_similar_templates(self):
        try:
            out = self._read_array(self._find_path('similar_templates.npy'))
            out = np.atleast_2d(out)
            assert out.ndim == 2
            return out
        except IOError:
            return np.zeros((self.n_templates, self.n_templates))

    def _load_templates(self):
        logger.debug("Loading templates.")

        # Sparse structure: regular array with col indices.
        try:
            path = self._find_path('templates.npy', 'clusters.templateWaveforms.npy')
            data = self._read_array(path, mmap_mode='r')
            data = np.atleast_3d(data)
            assert data.ndim == 3
            assert data.dtype in (np.float32, np.float64)
            n_templates, n_samples, n_channels_loc = data.shape
        except IOError:
            return

        try:
            # WARNING: KS2 saves templates_ind.npy (with an s), and note template_ind.npy,
            # so that file is not taken into account here.
            # That means templates.npy is considered as a dense array.
            # Proper fix would be to save templates.npy as a true sparse array, with proper
            # template_ind.npy (without an s).
            path = self._find_path('template_ind.npy')
            cols = self._read_array(path)
            cols = np.atleast_2d(cols)
            assert cols.ndim == 2
            logger.debug("Templates are sparse.")
            assert cols.shape == (n_templates, n_channels_loc)
        except IOError:
            logger.debug("Templates are dense.")
            cols = None

        return Bunch(data=data, cols=cols)

    def _load_wm(self):
        logger.debug("Loading the whitening matrix.")
        out = self._read_array(self._find_path('whitening_mat.npy'))
        out = np.atleast_2d(out)
        assert out.ndim == 2
        return out

    def _load_wmi(self):  # pragma: no cover
        logger.debug("Loading the inverse of the whitening matrix.")
        out = self._read_array(self._find_path('whitening_mat_inv.npy'))
        out = np.atleast_2d(out)
        assert out.ndim == 2
        return out

    def _compute_wmi(self, wm):
        logger.debug("Inversing the whitening matrix %s.", wm.shape)
        try:
            wmi = np.linalg.inv(wm)
        except np.linalg.LinAlgError as e:  # pragma: no cover
            raise ValueError(
                "Error when inverting the whitening matrix: %s.", e)
        self._write_array(self.dir_path / 'whitening_mat_inv.npy', wmi)
        return wmi

    def _unwhiten(self, x, channel_ids=None):
        mat = self.wmi
        if channel_ids is not None:
            mat = mat[np.ix_(channel_ids, channel_ids)]
            assert mat.shape == (len(channel_ids),) * 2
        assert x.shape[1] == mat.shape[0]
        return np.dot(np.ascontiguousarray(x),
                      np.ascontiguousarray(mat))

    def _load_features(self):

        # Sparse structure: regular array with row and col indices.
        try:
            logger.debug("Loading features.")
            data = self._read_array(
                self._find_path('pc_features.npy'))
            if data.ndim == 2:  # pragma: no cover
                # Deal with npcs = 1.
                data = data.reshape(data.shape + (1,))
            assert data.ndim == 3
            assert data.dtype in (np.float32, np.float64)
            data = data.transpose((0, 2, 1))
            n_spikes, n_channels_loc, n_pcs = data.shape
        except IOError:
            return

        try:
            cols = self._read_array(self._find_path('pc_feature_ind.npy'))
            logger.debug("Features are sparse.")
            if cols.ndim == 1:  # pragma: no cover
                # Deal with npcs = 1.
                cols = cols.reshape(cols.shape + (1,))
            assert cols.ndim == 2
            assert cols.shape == (self.n_templates, n_channels_loc)
        except IOError:
            logger.debug("Features are dense.")
            cols = None

        try:
            rows = self._read_array(self._find_path('pc_feature_spike_ids.npy'))
            assert rows.shape == (n_spikes,)
        except IOError:
            rows = None

        return Bunch(data=data, cols=cols, rows=rows)

    def _load_template_features(self):

        # Sparse structure: regular array with row and col indices.
        try:
            logger.debug("Loading template features.")
            data = self._read_array(self._find_path('template_features.npy'), mmap_mode='r')
            assert data.dtype in (np.float32, np.float64)
            assert data.ndim == 2
            n_spikes, n_channels_loc = data.shape
        except IOError:
            return

        try:
            cols = self._read_array(self._find_path('template_feature_ind.npy'))
            logger.debug("Template features are sparse.")
            assert cols.shape == (self.n_templates, n_channels_loc)
        except IOError:
            cols = None
            logger.debug("Template features are dense.")

        try:
            rows = self._read_array(self._find_path('template_feature_spike_ids.npy'))
            assert rows.shape == (n_spikes,)
        except IOError:
            rows = None

        return Bunch(data=data, cols=cols, rows=rows)

    #--------------------------------------------------------------------------
    # Internal data access methods
    #--------------------------------------------------------------------------

    def _find_best_channels(self, template):
        """Find the best channels for a given template."""
        # Compute the template amplitude on each channel.
        assert template.ndim == 2  # shape: (n_samples, n_channels)
        amplitude = template.max(axis=0) - template.min(axis=0)
        assert amplitude.ndim == 1  # shape: (n_channels,)
        # Find the peak channel.
        best_channel = np.argmax(amplitude)
        max_amp = amplitude[best_channel]
        # Find the channels X% peak.
        peak_channels = np.nonzero(amplitude >= self.amplitude_threshold * max_amp)[0]
        # Find N closest channels.
        close_channels = get_closest_channels(
            self.channel_positions, best_channel, self.n_closest_channels)
        assert best_channel in close_channels
        # Restrict to the channels belonging to the best channel's shank.
        if self.channel_shanks is not None:
            shank = self.channel_shanks[best_channel]  # shank of best channel
            channels_on_shank = np.nonzero(self.channel_shanks == shank)[0]
            close_channels = np.intersect1d(close_channels, channels_on_shank)
        # Keep the intersection.
        channel_ids = np.intersect1d(peak_channels, close_channels)
        # Order the channels by decreasing amplitude.
        order = np.argsort(amplitude[channel_ids])[::-1]
        channel_ids = channel_ids[order]
        amplitude = amplitude[order]
        assert best_channel in channel_ids
        assert amplitude.shape == (len(channel_ids),)
        return channel_ids, amplitude, best_channel

    def _get_template_dense(self, template_id, channel_ids=None):
        """Return data for one template."""
        template_w = self.sparse_templates.data[template_id, ...]
        template = self._unwhiten(template_w).astype(np.float32)
        assert template.ndim == 2
        channel_ids_, amplitude, best_channel = self._find_best_channels(template)
        channel_ids = channel_ids if channel_ids is not None else channel_ids_
        template = template[:, channel_ids]
        assert template.ndim == 2
        assert template.shape[1] == channel_ids.shape[0]
        return Bunch(
            template=template,
            amplitude=amplitude,
            best_channel=best_channel,
            channel_ids=channel_ids,
        )

    def _get_template_sparse(self, template_id):
        data, cols = self.sparse_templates.data, self.sparse_templates.cols
        assert cols is not None
        template_w, channel_ids = data[template_id], cols[template_id]

        # Remove unused channels = -1.
        used = channel_ids != -1
        template_w = template_w[:, used]
        channel_ids = channel_ids[used]
        channel_ids = channel_ids.astype(np.uint32)

        # Unwhiten.
        template = self._unwhiten(template_w, channel_ids=channel_ids)
        template = template.astype(np.float32)
        assert template.ndim == 2
        assert template.shape[1] == len(channel_ids)
        # Compute the amplitude and the channel with max amplitude.
        amplitude = template.max(axis=0) - template.min(axis=0)
        best_channel = channel_ids[np.argmax(amplitude)]
        return Bunch(
            template=template,
            amplitude=amplitude,
            best_channel=best_channel,
            channel_ids=channel_ids,
        )

    #--------------------------------------------------------------------------
    # Data access methods
    #--------------------------------------------------------------------------

    def get_template(self, template_id, channel_ids=None):
        """Get data about a template."""
        if self.sparse_templates.cols is not None:
            return self._get_template_sparse(template_id)
        else:
            return self._get_template_dense(template_id, channel_ids=channel_ids)

    def get_waveforms(self, spike_ids, channel_ids=None):
        """Return spike waveforms on specified channels."""
        if self.waveform_loader is None:
            return
        out = self.waveform_loader.get(spike_ids, channel_ids)
        assert out.dtype in (np.float32, np.float64)
        assert out.shape[0] == len(spike_ids)
        if channel_ids is not None:
            assert out.shape[2] == len(channel_ids)
        return out

    def get_features(self, spike_ids, channel_ids):
        """Return sparse features for given spikes."""
        sf = self.sparse_features
        if sf is None:
            return
        _, n_channels_loc, n_pcs = sf.data.shape
        ns = len(spike_ids)
        nc = len(channel_ids)

        # Initialize the output array.
        features = np.empty((ns, n_channels_loc, n_pcs))
        features[:] = np.NAN

        if sf.rows is not None:
            s = np.intersect1d(spike_ids, sf.rows)
            # Relative indices of the spikes in the self.features_spike_ids
            # array, necessary to load features from all_features which only
            # contains the subset of the spikes.
            rows = _index_of(s, sf.rows)
            # Relative indices of the non-null rows in the output features
            # array.
            rows_out = _index_of(s, spike_ids)
        else:
            rows = spike_ids
            rows_out = slice(None, None, None)
        features[rows_out, ...] = sf.data[rows]

        if sf.cols is not None:
            assert sf.cols.shape[1] == n_channels_loc
            cols = sf.cols[self.spike_templates[spike_ids]]
        else:
            cols = np.tile(np.arange(n_channels_loc), (ns, 1))
        features = from_sparse(features, cols, channel_ids)

        assert features.shape == (ns, nc, n_pcs)
        return features

    def get_template_features(self, spike_ids):
        """Return sparse template features for given spikes."""
        tf = self.sparse_template_features
        if tf is None:
            return
        _, n_templates_loc = tf.data.shape
        ns = len(spike_ids)

        if tf.rows is not None:
            spike_ids = np.intersect1d(spike_ids, tf.rows)
            # Relative indices of the spikes in the self.features_spike_ids
            # array, necessary to load features from all_features which only
            # contains the subset of the spikes.
            rows = _index_of(spike_ids, tf.rows)
        else:
            rows = spike_ids
        template_features = tf.data[rows]

        if tf.cols is not None:
            assert tf.cols.shape[1] == n_templates_loc
            cols = tf.cols[self.spike_templates[spike_ids]]
        else:
            cols = np.tile(np.arange(n_templates_loc), (len(spike_ids), 1))
        template_features = from_sparse(template_features, cols, np.arange(self.n_templates))

        assert template_features.shape[0] == ns
        return template_features

    #--------------------------------------------------------------------------
    # Internal helper methods for public high-level methods
    #--------------------------------------------------------------------------

    def _get_template_from_spikes(self, spike_ids):
        """Get the main template from a set of spikes."""
        # We get the original template ids for the spikes.
        st = self.spike_templates[spike_ids]
        # We find the template with the largest number of spikes from the spike selection.
        template_ids, counts = np.unique(st, return_counts=True)
        ind = np.argmax(counts)
        template_id = template_ids[ind]
        # We load the template.
        template = self.get_template(template_id)
        # We return the template.
        return template

    #--------------------------------------------------------------------------
    # Public high-level methods
    #--------------------------------------------------------------------------

    def describe(self):
        """Display basic information about the dataset."""
        def _print(name, value):
            print("{0: <24}{1}".format(name, value))

        _print('Data files', ', '.join(map(str, self.dat_path)))
        _print('Directory', self.dir_path)
        _print('Duration', '{:.1f}s'.format(self.duration))
        _print('Sample rate', '{:.1f} kHz'.format(self.sample_rate / 1000.))
        _print('Data type', self.dtype)
        _print('# of channels', self.n_channels)
        _print('# of channels (raw)', self.n_channels_dat)
        _print('# of templates', self.n_templates)
        _print('# of spikes', "{:,}".format(self.n_spikes))

    def get_template_spikes(self, template_id):
        """Return the spike ids that belong to a given template."""
        return _spikes_in_clusters(self.spike_templates, [template_id])

    def get_cluster_spikes(self, cluster_id):
        """Return the spike ids that belong to a given template."""
        return _spikes_in_clusters(self.spike_clusters, [cluster_id])

    def get_template_channels(self, template_id):
        """Return the most relevant channels of a template."""
        template = self.get_template(template_id)
        return template.channel_ids

    def get_cluster_channels(self, cluster_id):
        """Return the most relevant channels of a cluster."""
        spike_ids = self.get_cluster_spikes(cluster_id)
        return self._get_template_from_spikes(spike_ids).channel_ids

    def get_template_waveforms(self, template_id):
        """Return the waveforms of a template on the most relevant channels."""
        template = self.get_template(template_id)
        return template.template

    def get_template_spike_waveforms(self, template_id):
        """Return all spike waveforms of a template, on the most relevant channels."""
        spike_ids = self.get_template_spikes(template_id)
        channel_ids = self.get_template_channels(template_id)
        return self.get_waveforms(spike_ids, channel_ids)

    def get_cluster_spike_waveforms(self, cluster_id):
        """Return all spike waveforms of a cluster, on the most relevant channels."""
        spike_ids = self.get_cluster_spikes(cluster_id)
        channel_ids = self.get_cluster_channels(cluster_id)
        return self.get_waveforms(spike_ids, channel_ids)

    #--------------------------------------------------------------------------
    # Saving methods
    #--------------------------------------------------------------------------

    def save_metadata(self, name, values):
        """Save a dictionary {cluster_id: value} with cluster metadata in
        a TSV file."""
        path = self.dir_path / ('cluster_%s.tsv' % name)
        logger.debug("Save cluster metadata to `%s`.", path)
        # Remove empty values.
        save_metadata(
            path, name, {c: v for c, v in values.items() if v is not None})

    def save_spike_clusters(self, spike_clusters):
        """Save the spike clusters."""
        path = self._find_path('spike_clusters.npy', 'spikes.clusters.npy', multiple_ok=False)
        logger.debug("Save spike clusters to `%s`.", path)
        np.save(path, spike_clusters)

    def save_mean_waveforms(self, mean_waveforms):
        """Save the mean waveforms as a single array."""
        path = self.dir_path / 'clusters.meanWaveforms.npy'
        n_clusters = len(mean_waveforms)
        out = np.zeros((n_clusters, self.n_samples_waveforms, self.n_channels))
        for i, cluster_id in enumerate(sorted(mean_waveforms)):
            b = mean_waveforms[cluster_id]
            if b.data is not None:
                out[i, :, b.channel_ids] = b.data[0, ...].T
        logger.debug("Save mean waveforms to `%s`.", path)
        np.save(path, out)

    def close(self):
        """Close all memmapped files."""
        for k, v in sorted(self.__dict__.items(), key=itemgetter(0)):
            _close_memmap(k, v)


def _make_abs_path(p, dir_path):
    p = Path(p)
    if not op.isabs(p):
        p = dir_path / p
    if not p.exists():
        logger.warning("File %s does not exist.", p)
    return p


def get_template_params(params_path):
    """Get a dictionary of parameters from a `params.py` file."""
    params_path = Path(params_path)

    params = read_python(params_path)
    params['dtype'] = np.dtype(params['dtype'])

    if 'dir_path' not in params:
        params['dir_path'] = params_path.parent
    params['dir_path'] = Path(params['dir_path'])
    assert params['dir_path'].is_dir()
    assert params['dir_path'].exists()

    if isinstance(params['dat_path'], str):
        params['dat_path'] = [params['dat_path']]
    params['dat_path'] = [_make_abs_path(_, params['dir_path']) for _ in params['dat_path']]
    return params


def load_model(params_path):
    """Return a TemplateModel instance from a path to a `params.py` file."""
    return TemplateModel(**get_template_params(params_path))



# -*- coding: utf-8 -*-

"""Test ALF dataset generation."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import os
import os.path as op
from pathlib import Path
import shutil
from pytest import fixture, raises

import numpy as np
import numpy.random as nr

from phylib.utils._misc import _write_tsv_simple
from ..alf import _FILE_RENAMES, _load, EphysAlfCreator
from ..model import TemplateModel


#------------------------------------------------------------------------------
# Fixture
#------------------------------------------------------------------------------

class Dataset(object):
    def __init__(self, tempdir):
        self.tmp_dir = tempdir
        p = Path(self.tmp_dir)
        self.ns = 100
        self.nc = 10
        self.nt = 5
        self.ncd = 1000
        np.save(p / 'spike_times.npy', .01 * np.cumsum(nr.exponential(size=self.ns)))
        np.save(p / 'spike_clusters.npy', nr.randint(low=10, high=10 + self.nt, size=self.ns))
        shutil.copy(p / 'spike_clusters.npy', p / 'spike_templates.npy')
        np.save(p / 'amplitudes.npy', nr.uniform(low=0.5, high=1.5, size=self.ns))
        np.save(p / 'channel_positions.npy', np.c_[np.arange(self.nc), np.zeros(self.nc)])
        np.save(p / 'templates.npy', np.random.normal(size=(self.nt, 50, self.nc)))
        np.save(p / 'similar_templates.npy', np.tile(np.arange(self.nt), (self.nt, 1)))
        np.save(p / 'channel_map.npy', np.c_[np.arange(self.nc)])
        np.save(p / 'channel_probe.npy', np.zeros(self.nc))
        _write_tsv_simple(p / 'cluster_group.tsv', 'group', {2: 'good', 3: 'mua', 5: 'noise'})
        _write_tsv_simple(p / 'cluster_Amplitude.tsv', field_name='Amplitude',
                          data={str(n): np.random.rand() * 120 for n in np.arange(self.nt)})
        with open(p / 'probes.description.txt', 'w+') as fid:
            fid.writelines(['label\n'])

        # Raw data
        self.dat_path = p / 'rawdata.npy'
        np.save(self.dat_path, np.random.normal(size=(self.ncd, self.nc)))

        # LFP data.
        lfdata = (100 * np.random.normal(size=(1000, self.nc))).astype(np.int16)
        with (p / 'mydata.lf.bin').open('wb') as f:
            lfdata.tofile(f)

        self.files = os.listdir(self.tmp_dir)

    def _load(self, fn):
        p = Path(self.tmp_dir)
        return _load(p / fn)


@fixture
def dataset(tempdir):
    return Dataset(tempdir)


def test_ephys_1(dataset):
    assert dataset._load('spike_times.npy').shape == (dataset.ns,)
    assert dataset._load('spike_clusters.npy').shape == (dataset.ns,)
    assert dataset._load('amplitudes.npy').shape == (dataset.ns,)
    assert dataset._load('channel_positions.npy').shape == (dataset.nc, 2)
    assert dataset._load('templates.npy').shape == (dataset.nt, 50, dataset.nc)
    assert dataset._load('channel_map.npy').shape == (dataset.nc, 1)
    assert dataset._load('channel_probe.npy').shape == (dataset.nc,)
    assert len(dataset._load('cluster_group.tsv')) == 3
    assert dataset._load('rawdata.npy').shape == (1000, dataset.nc)
    assert dataset._load('mydata.lf.bin').shape == (1000 * dataset.nc,)


def test_creator(dataset):
    _FILE_CREATES = (
        'spikes.times.npy',
        'clusters.peakChannel.npy',
        'clusters.waveformDuration.npy',
        'spikes.depths.npy',
        'clusters.depths.npy',
        'clusters.meanWaveforms.npy',
        # 'clusters.probes.npy',
    )
    path = Path(dataset.tmp_dir)
    out_path = path / 'alf'

    model = TemplateModel(
        dir_path=path, dat_path=dataset.dat_path, sample_rate=2000, n_channels_dat=dataset.nc)

    c = EphysAlfCreator(model)
    with raises(IOError):
        c.convert(dataset.tmp_dir)
    c.convert(out_path)

    # Check all renames.
    for old, new, _ in _FILE_RENAMES:
        if op.exists(str(path / old)):
            assert op.exists(str(out_path / new))

    for new in _FILE_CREATES:
        assert (out_path / new).exists()

    model.close()



# -*- coding: utf-8 -*-

"""Plotting utilities."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import logging

import numpy as np

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Common probe layouts
#------------------------------------------------------------------------------

def linear_positions(n_channels):
    """Linear channel positions along the vertical axis."""
    return np.c_[np.zeros(n_channels),
                 np.linspace(0., 1., n_channels)]


def staggered_positions(n_channels):
    """Generate channel positions for a staggered probe."""
    i = np.arange(n_channels - 1)
    x, y = (-1) ** i * (5 + i), 10 * (i + 1)
    pos = np.flipud(np.r_[np.zeros((1, 2)), np.c_[x, y]])
    return pos


#------------------------------------------------------------------------------
# Box positioning
#------------------------------------------------------------------------------

def range_transform(from_bounds, to_bounds, positions):
    """Transform for a rectangle to another."""
    from_bounds = np.asarray(from_bounds)
    to_bounds = np.asarray(to_bounds)
    positions = np.asarray(positions)

    f0 = from_bounds[..., :2]
    f1 = from_bounds[..., 2:]
    t0 = to_bounds[..., :2]
    t1 = to_bounds[..., 2:]

    d = (f1 - f0)
    d[d == 0] = 1

    out = positions.copy()
    out -= f0
    out *= (t1 - t0) / d
    out += t0
    return out


def _boxes_overlap(x0, y0, x1, y1):
    n = len(x0)
    overlap_matrix = ((x0 < x1.T) & (x1 > x0.T) & (y0 < y1.T) & (y1 > y0.T))
    overlap_matrix[np.arange(n), np.arange(n)] = False
    return np.any(overlap_matrix.ravel())


def _binary_search(f, xmin, xmax, eps=1e-9):
    """Return the largest x such f(x) is True."""
    middle = (xmax + xmin) / 2.
    while xmax - xmin > eps:
        assert xmin < xmax
        middle = (xmax + xmin) / 2.
        if f(xmax):
            return xmax
        if not f(xmin):
            return xmin
        if f(middle):
            xmin = middle
        else:
            xmax = middle
    return middle


def _get_box_size(x, y, ar=.5, margin=0):
    logger.log(5, "Get box size for %d points.", len(x))
    # Deal with degenerate x case.
    xmin, xmax = x.min(), x.max()
    if xmin == xmax:
        # If all positions are vertical, the width can be maximum.
        wmax = 1.
    else:
        wmax = xmax - xmin

    def f1(w):
        """Return true if the configuration with the current box size
        is non-overlapping."""
        # NOTE: w|h are the *half* width|height.
        h = w * ar  # fixed aspect ratio
        return not _boxes_overlap(x - w, y - h, x + w, y + h)

    # Find the largest box size leading to non-overlapping boxes.
    w = _binary_search(f1, 0, wmax)
    w = w * (1 - margin)  # margin
    # Clip the half-width.
    h = w * ar  # aspect ratio

    return w, h


def _get_boxes(pos, size=None, margin=0, keep_aspect_ratio=True):
    """Generate non-overlapping boxes in NDC from a set of positions."""

    # Margin.
    a = margin

    # Get x, y.
    pos = np.asarray(pos, dtype=np.float64)
    x, y = pos.T
    x = x[:, np.newaxis]
    y = y[:, np.newaxis]

    w, h = size if size is not None else _get_box_size(x, y, margin=margin)

    x0, y0 = x - w, y - h
    x1, y1 = x + w, y + h

    # Renormalize the whole thing by keeping the aspect ratio.
    x0min, y0min, x1max, y1max = x0.min(), y0.min(), x1.max(), y1.max()
    if not keep_aspect_ratio:
        b = (x0min, y0min, x1max, y1max)
    else:
        dx = x1max - x0min
        dy = y1max - y0min
        if dx > dy:
            b = (x0min, (y1max + y0min) / 2. - dx / 2.,
                 x1max, (y1max + y0min) / 2. + dx / 2.)
        else:
            b = ((x1max + x0min) / 2. - dy / 2., y0min,
                 (x1max + x0min) / 2. + dy / 2., y1max)
    fb = np.asarray(b)
    tb = np.asarray((-1 + a, -1 + a, 1 - a, 1 - a))
    return np.c_[
        range_transform(fb, tb, np.c_[x0, y0]),
        range_transform(fb, tb, np.c_[x1, y1])
    ]


def _get_box_pos_size(box_bounds):
    box_bounds = np.asarray(box_bounds)
    x0, y0, x1, y1 = box_bounds.T
    w = (x1 - x0) * .5
    h = (y1 - y0) * .5
    x = (x0 + x1) * .5
    y = (y0 + y1) * .5
    return np.c_[x, y], (w.mean(), h.mean())


def _check_data_bounds(data_bounds):
    """Check that a data bounds is a valid."""
    assert data_bounds.ndim == 2
    assert data_bounds.shape[1] == 4
    assert np.all(data_bounds[:, 0] < data_bounds[:, 2])
    assert np.all(data_bounds[:, 1] < data_bounds[:, 3])


def _get_data_bounds(data_bounds, pos=None, length=None):
    """"Prepare data bounds, possibly using min/max of the data."""
    if data_bounds is None or (isinstance(data_bounds, str) and data_bounds == 'auto'):
        if pos is not None and len(pos):
            m, M = pos.min(axis=0), pos.max(axis=0)
            data_bounds = [m[0], m[1], M[0], M[1]]
        else:
            data_bounds = [-1, -1, 1, 1]
    data_bounds = np.atleast_2d(data_bounds)

    ind_x = data_bounds[:, 0] == data_bounds[:, 2]
    ind_y = data_bounds[:, 1] == data_bounds[:, 3]
    if np.sum(ind_x):
        data_bounds[ind_x, 0] -= 1
        data_bounds[ind_x, 2] += 1
    if np.sum(ind_y):
        data_bounds[ind_y, 1] -= 1
        data_bounds[ind_y, 3] += 1

    # Extend the data_bounds if needed.
    if length is None:
        length = pos.shape[0] if pos is not None else 1
    if data_bounds.shape[0] == 1:
        data_bounds = np.tile(data_bounds, (length, 1))

    # Check the shape of data_bounds.
    assert data_bounds.shape == (length, 4)

    _check_data_bounds(data_bounds)
    return data_bounds



# -*- coding: utf-8 -*-

"""Color routines."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import colorcet as cc
import logging

from phylib.utils import Bunch
from phylib.io.array import _index_of

import numpy as np
from numpy.random import uniform
from matplotlib.colors import hsv_to_rgb, rgb_to_hsv

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Random colors
#------------------------------------------------------------------------------

def _random_color(h_range=(0., 1.), s_range=(.5, 1.), v_range=(.5, 1.)):
    """Generate a random RGB color."""
    h, s, v = uniform(*h_range), uniform(*s_range), uniform(*v_range)
    r, g, b = hsv_to_rgb(np.array([[[h, s, v]]])).flat
    return r, g, b


def _is_bright(rgb):
    """Return whether a RGB color is bright or not.
    see https://stackoverflow.com/a/3943023/1595060
    """
    L = 0
    for c, coeff in zip(rgb, (0.2126, 0.7152, 0.0722)):
        if c <= 0.03928:
            c = c / 12.92
        else:
            c = ((c + 0.055) / 1.055) ** 2.4
        L += c * coeff
    if (L + 0.05) / (0.0 + 0.05) > (1.0 + 0.05) / (L + 0.05):
        return True


def _random_bright_color():
    """Generate a random bright color."""
    rgb = _random_color()
    while not _is_bright(rgb):
        rgb = _random_color()
    return rgb


def _hex_to_triplet(h):
    """Convert an hexadecimal color to a triplet of int8 integers."""
    if h.startswith('#'):
        h = h[1:]
    return tuple(int(h[i:i + 2], 16) for i in (0, 2, 4))


def _override_hsv(rgb, h=None, s=None, v=None):
    h_, s_, v_ = rgb_to_hsv(np.array([[rgb]])).flat
    h = h if h is not None else h_
    s = s if s is not None else s_
    v = v if v is not None else v_
    r, g, b = hsv_to_rgb(np.array([[[h, s, v]]])).flat
    return r, g, b


#------------------------------------------------------------------------------
# Colormap utilities
#------------------------------------------------------------------------------

def _continuous_colormap(colormap, values, vmin=None, vmax=None):
    """Convert values into colors given a specified continuous colormap."""
    assert values is not None
    assert colormap.shape[1] == 3
    n = colormap.shape[0]
    vmin = vmin if vmin is not None else values.min()
    vmax = vmax if vmax is not None else values.max()
    assert vmin is not None
    assert vmax is not None
    denom = vmax - vmin
    denom = denom if denom != 0 else 1
    # NOTE: clipping is necessary when a view using color selector (like the raster view)
    # is updated right after a clustering update, but before the vmax had a chance to
    # be updated.
    i = np.clip(np.round((n - 1) * (values - vmin) / denom).astype(np.int32), 0, n - 1)
    return colormap[i, :]


def _categorical_colormap(colormap, values, vmin=None, vmax=None):
    """Convert values into colors given a specified categorical colormap."""
    assert np.issubdtype(values.dtype, np.integer)
    assert colormap.shape[1] == 3
    n = colormap.shape[0]
    if vmin is None and vmax is None:
        # Find unique values and keep the order.
        _, idx = np.unique(values, return_index=True)
        lookup = values[np.sort(idx)]
        x = _index_of(values, lookup)
    else:
        x = values
    return colormap[x % n, :]


#------------------------------------------------------------------------------
# Colormaps
#------------------------------------------------------------------------------

# Default color map for the selected clusters.
# see https://colorcet.pyviz.org/user_guide/Categorical.html
def _make_default_colormap():
    """Return the default colormap, with custom first colors."""
    colormap = np.array(cc.glasbey_bw_minc_20_minl_30)
    # Reorder first colors.
    colormap[[0, 1, 2, 3, 4, 5]] = colormap[[3, 0, 4, 5, 2, 1]]
    # Replace first two colors.
    colormap[0] = [0.03137, 0.5725, 0.9882]
    colormap[1] = [1.0000, 0.0078, 0.0078]
    return colormap


def _make_cluster_group_colormap():
    """Return cluster group colormap."""
    # Rows are sorted by increasing alphabetical order (all lowercase).
    return np.array([
        [0.75, 0.75, 0.75],  # '' (None = '' = unsorted)
        [0.5254, 0.8196, 0.42745],  # good
        [0.5, 0.5, 0.5],  # mua
        [0.4, 0.4, 0.4],  # noise
    ])


"""Built-in colormaps."""
colormaps = Bunch(
    default=_make_default_colormap(),
    cluster_group=_make_cluster_group_colormap(),
    categorical=np.array(cc.glasbey_bw_minc_20_minl_30),
    rainbow=np.array(cc.rainbow_bgyr_35_85_c73),
    linear=np.array(cc.linear_wyor_100_45_c55),
    diverging=np.array(cc.diverging_linear_bjy_30_90_c45),
)


def selected_cluster_color(i, alpha=1.):
    """Return the color, as a 4-tuple, of the i-th selected cluster."""
    return add_alpha(tuple(colormaps.default[i % len(colormaps.default)]), alpha=alpha)


def spike_colors(spike_clusters, cluster_ids):
    """Return the colors of spikes according to the index of their cluster within `cluster_ids`.

    Parameters
    ----------

    spike_clusters : array-like
        The spike-cluster assignments.
    cluster_ids : array-like
        The set of unique selected cluster ids appearing in spike_clusters, in a given order

    Returns
    -------

    spike_colors : array-like
        For each spike, the RGBA color (in [0,1]) depending on the index of the cluster within
        `cluster_ids`.

    """
    spike_clusters_idx = _index_of(spike_clusters, cluster_ids)
    return add_alpha(colormaps.default[np.mod(spike_clusters_idx, colormaps.default.shape[0])])


def _add_selected_clusters_colors(selected_clusters, cluster_ids, cluster_colors=None):
    """Take an array with colors of clusters as input, and add colors of selected clusters."""
    # Find the index of the selected clusters within the self.cluster_ids.
    clu_idx = _index_of(selected_clusters, cluster_ids)
    # Get the colors of the selected clusters.
    colormap = _categorical_colormap(colormaps.default, clu_idx)
    # Inject those colors in cluster_colors.
    cluster_colors[clu_idx] = add_alpha(colormap, 1)
    return cluster_colors


#------------------------------------------------------------------------------
# Cluster color selector
#------------------------------------------------------------------------------

def add_alpha(c, alpha=1.):
    """Add an alpha channel to an RGB color.

    Parameters
    ----------

    c : array-like (2D, shape[1] == 3) or 3-tuple
    alpha : float

    """
    if isinstance(c, (tuple,)):
        if len(c) == 4:
            c = c[:3]
        return c + (alpha,)
    elif isinstance(c, np.ndarray):
        if c.shape[-1] == 4:
            c = c[..., :3]
        assert c.shape[-1] == 3
        out = np.concatenate([c, alpha * np.ones((c.shape[:-1] + (1,)))], axis=-1)
        assert out.ndim == c.ndim
        assert out.shape[-1] == c.shape[-1] + 1
        return out
    raise ValueError("Unknown value given in add_alpha().")


def _categorize(values):
    """Categorize a lsit of values by replacing strings and None values by integers."""
    if any(isinstance(v, str) or v is None for v in values):
        # HACK: replace None by empty string to avoid error when sorting the unique values.
        values = [str(v).lower() if v is not None else '' for v in values]
        uv = sorted(set(values))
        values = [uv.index(v) for v in values]
    return values


class ClusterColorSelector(object):
    """Assign a color to clusters depending on cluster labels or metrics."""
    _color_field = 'cluster'
    _colormap = colormaps.categorical
    _categorical = True
    _logarithmic = False

    def __init__(
            self, cluster_meta=None, cluster_metrics=None, color_field=None,
            colormap=None, categorical=None, logarithmic=None, cluster_ids=None):
        self.cluster_ids = None
        self.cluster_meta = cluster_meta or None
        self.cluster_meta_fields = cluster_meta.fields if cluster_meta else ()
        self.cluster_metrics = cluster_metrics or {}
        # Used to initialize the value range for the clusters.
        assert cluster_ids is not None
        self.cluster_ids = cluster_ids
        # self._colormap = colormap if colormap is not None else self._colormap
        self.set_color_mapping(
            color_field=color_field, colormap=colormap,
            categorical=categorical, logarithmic=logarithmic)

    @property
    def state(self):
        """Colormap state. This is a Bunch with the following keys: color_field, colormap,
        categorical, logarithmic."""
        colormap_name = None
        # Find the colormap name from the colormap array.
        for cname, arr in colormaps.items():
            if self._colormap.shape == arr.shape and np.allclose(self._colormap, arr):
                colormap_name = cname
                break
        return Bunch(
            color_field=self._color_field,
            colormap=colormap_name,
            categorical=self._categorical,
            logarithmic=self._logarithmic,
        )

    def set_state(self, state):
        """Set the colormap state."""
        self.set_color_mapping(
            color_field=state.color_field, colormap=state.colormap,
            categorical=state.categorical, logarithmic=state.logarithmic)

    def set_color_mapping(
            self, color_field=None, colormap=None, categorical=None, logarithmic=None):
        """Set the field used to choose the cluster colors, and the associated colormap.

        Parameters
        ----------

        color_field : str
            Name of the cluster metrics or label to use for the color.
        colormap : array-like
            A `(N, 3)` array with the colormaps colors
        categorical : boolean
            Whether the colormap is categorical (one value = one color) or continuous (values
            are continuously mapped from their initial interval to the colors).
        logarithmic : boolean
            Whether to use a logarithmic transform for the mapping.

        """
        if isinstance(colormap, str):
            colormap = colormaps[colormap]
        self._colormap = colormap if colormap is not None else self._colormap
        self._color_field = color_field or self._color_field
        self._categorical = categorical if categorical is not None else self._categorical
        self._logarithmic = logarithmic if logarithmic is not None else self._logarithmic
        # Recompute the value range.
        self.set_cluster_ids(self.cluster_ids)

    def set_cluster_ids(self, cluster_ids):
        """Precompute the value range for all clusters."""
        self.cluster_ids = cluster_ids
        values = self.get_values(self.cluster_ids)
        if values is not None:
            self.vmin, self.vmax = values.min(), values.max()

    def map(self, values):
        """Convert values to colors using the selected colormap.

        Parameters
        ----------

        values : array-like (1D)

        Returns
        -------

        colors : array-like (2D, shape[1] == 3)

        """
        if self._logarithmic:
            assert np.all(values > 0)
            values = np.log(values)
            vmin, vmax = np.log(self.vmin), np.log(self.vmax)
        else:
            vmin, vmax = self.vmin, self.vmax
        assert values is not None
        # Use categorical or continuous colormap depending on the categorical option.
        f = (_categorical_colormap
             if self._categorical and np.issubdtype(values.dtype, np.integer)
             else _continuous_colormap)
        return f(self._colormap, values, vmin=vmin, vmax=vmax)

    def _get_cluster_value(self, cluster_id):
        """Return the field value for a given cluster."""
        field = self._color_field
        if field == 'cluster':
            return cluster_id
        elif field in self.cluster_meta_fields:
            return self.cluster_meta.get(field, cluster_id)
        elif field in self.cluster_metrics:
            return self.cluster_metrics.get(field, lambda cl: None)(cluster_id)
        logger.warning("The field %s is not an existing cluster label or metrics.", field)
        return 0

    def get(self, cluster_id, alpha=None):
        """Return the RGBA color of a single cluster."""
        assert self.cluster_ids is not None
        assert self._colormap is not None
        val = self._get_cluster_value(cluster_id)
        col = tuple(self.map(np.array(_categorize([val])))[0].tolist())
        return add_alpha(col, alpha=alpha)

    def get_values(self, cluster_ids):
        """Get the values of clusters for the selected color field.."""
        values = [self._get_cluster_value(cluster_id) for cluster_id in cluster_ids]
        values = _categorize(values)
        return np.array(values)

    def get_colors(self, cluster_ids, alpha=1.):
        """Return the RGBA colors of some clusters."""
        values = self.get_values(cluster_ids)
        assert values is not None
        assert len(values) == len(cluster_ids)
        return add_alpha(self.map(values), alpha=alpha)



# -*- coding: utf-8 -*-

"""Test colors."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import colorcet as cc
import numpy as np
from numpy.testing import assert_almost_equal as ae

from phylib.utils import Bunch
from ..color import (
    _is_bright, _random_bright_color, spike_colors, add_alpha, selected_cluster_color,
    _override_hsv,
    _hex_to_triplet, _continuous_colormap, _categorical_colormap, ClusterColorSelector,
    colormaps, _add_selected_clusters_colors)


#------------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------------

def test_random_color():
    for _ in range(20):
        assert _is_bright(_random_bright_color())


def test_hex_to_triplet():
    assert _hex_to_triplet('#0123ab')


def test_add_alpha():
    assert add_alpha((0, .5, 1), .75) == (0, .5, 1, .75)
    assert add_alpha(np.random.rand(5, 3), .5).shape == (5, 4)

    assert add_alpha((0, .5, 1, .1), .75) == (0, .5, 1, .75)
    assert add_alpha(np.random.rand(5, 4), .5).shape == (5, 4)


def test_override_hsv():
    assert _override_hsv((.1, .9, .5), h=1, s=0, v=1) == (1, 1, 1)


def test_selected_cluster_color():
    c = selected_cluster_color(0)
    assert isinstance(c, tuple)
    assert len(c) == 4


def test_colormaps():
    colormap = np.array(cc.glasbey_bw_minc_20_minl_30)
    values = np.random.randint(10, 20, size=100)
    colors = _categorical_colormap(colormap, values)
    assert colors.shape == (100, 3)

    colormap = np.array(cc.rainbow_bgyr_35_85_c73)
    values = np.linspace(0, 1, 100)
    colors = _continuous_colormap(colormap, values)
    assert colors.shape == (100, 3)


def test_spike_colors():
    spike_clusters = [1, 0, 0, 3]
    cluster_ids = [0, 1, 2, 3]
    colors = spike_colors(spike_clusters, cluster_ids)
    assert colors.shape == (4, 4)
    ae(colors[1], colors[2])


def test_cluster_color_selector():
    # Mock ClusterMeta instance, with 'fields' property and get(field, cluster) function.
    cluster_meta = Bunch(fields=('label',), get=lambda f, cl: {1: 10, 2: 20, 3: 30}[cl])
    cluster_metrics = {'quality': lambda c: c * .1}
    cluster_ids = [1, 2, 3]
    c = ClusterColorSelector(
        cluster_meta=cluster_meta,
        cluster_metrics=cluster_metrics,
        cluster_ids=cluster_ids,
    )

    assert len(c.get(1, alpha=.5)) == 4
    ae(c.get_values([None, 0]), np.arange(2))

    for color_field, colormap in (
            ('label', 'linear'),
            ('quality', 'rainbow'),
            ('cluster', 'categorical'),
            ('nonexisting', 'diverging')):
        c.set_color_mapping(color_field=color_field, colormap=colormap)
        colors = c.get_colors(cluster_ids)
        assert colors.shape == (3, 4)

    # Get the state.
    assert c.state == {
        'color_field': 'nonexisting', 'colormap': 'diverging',
        'categorical': True, 'logarithmic': False}

    # Set the state.
    state = Bunch(c.state)
    state.color_field = 'label'
    state.colormap = colormaps.rainbow
    state.categorical = False
    c.set_state(state)

    # Check that the state was correctly set.
    assert c._color_field == 'label'
    ae(c._colormap, colormaps.rainbow)
    assert c._categorical is False


def test_cluster_color_group():
    # Mock ClusterMeta instance, with 'fields' property and get(field, cluster) function.
    cluster_meta = Bunch(fields=('group',), get=lambda f, cl: {1: None, 2: 'mua', 3: 'good'}[cl])
    cluster_ids = [1, 2, 3]
    c = ClusterColorSelector(
        cluster_meta=cluster_meta,
        cluster_ids=cluster_ids,
    )

    c.set_color_mapping(color_field='group', colormap='cluster_group')
    colors = c.get_colors(cluster_ids)
    assert colors.shape == (3, 4)


def test_cluster_color_log():
    cluster_ids = [1, 2, 3]
    c = ClusterColorSelector(
        cluster_ids=cluster_ids,
    )

    c.set_color_mapping(logarithmic=True)
    colors = c.get_colors(cluster_ids)
    assert colors.shape == (3, 4)


def test_add_selected_clusters_colors():
    cluster_colors = np.tile(np.c_[np.arange(3)], (1, 3))
    cluster_colors = add_alpha(cluster_colors)
    cluster_colors_sel = _add_selected_clusters_colors([1], [0, 1, 3], cluster_colors)
    ae(cluster_colors_sel[[0]], add_alpha(np.zeros((1, 3))))
    ae(cluster_colors_sel[[2]], add_alpha(2 * np.ones((1, 3))))
    # Cluster at index 0 is selected, should be in blue.
    r, g, b, _ = cluster_colors_sel[1]
    assert b > g > r



# -*- coding: utf-8 -*-

"""Probe layout."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

from collections import defaultdict

import numpy as np

from phylib.io.array import _flatten
from phylib.utils.geometry import _get_boxes, range_transform
from phylib.utils.color import selected_cluster_color


#------------------------------------------------------------------------------
# Layout
#------------------------------------------------------------------------------

def _iter_channel(positions):
    """Iterate through the channel positions."""
    size = 100
    margin = 5
    boxes = _get_boxes(positions, keep_aspect_ratio=False)
    xmin, ymin = boxes[:, :2].min(axis=0)
    xmax, ymax = boxes[:, 2:].max(axis=0)
    x = boxes[:, [0, 2]].mean(axis=1)
    y = - boxes[:, [1, 3]].mean(axis=1)
    positions = np.c_[x, y]
    tr = [margin, margin, size - margin, size - margin]
    positions_tr = range_transform([-1, -1, 1, 1], tr, positions)
    for x, y in positions_tr:
        yield x, y


def _disk(x, y, r, c, t=0):
    """Return a SVG disc."""
    return ('<circle cx="%.5f%%" cy="%.5f%%" '
            'r="%d" fill="%s" '
            'transform="translate(%d, 0)"'
            ' />') % (x, y, r, c, t)


def _rgba(rgb, a=1.):
    r, g, b = rgb
    return 'rgba(%d, %d, %d, %.3f)' % (255 * r, 255 * g, 255 * b, a)


def _iter_disks(positions, cluster_channels=None):
    """Iterate through the SVG dics for all channels and clusters.

    positions: Nx2 array
    cluster_channels: {cluster: channels}

    """

    color_masked = _rgba((.5, .5, .5), 1.)
    color_umasked = lambda clu: _rgba(selected_cluster_color(clu)[:3], 1)

    size_masked = 5
    size_unmasked = 7

    # n_channels = positions.shape[0]
    n_clusters = len(cluster_channels)

    cluster_channels = cluster_channels or {}
    channel_ids = set(_flatten(cluster_channels.values()))
    is_masked = {ch: False for ch in channel_ids}

    # List of clusters per channel.
    clusters_per_channel = defaultdict(lambda: [])
    for clu, channels in cluster_channels.items():
        for channel in channels:
            clusters_per_channel[channel].append(clu)

    # Enumerate the discs for each channel.
    for channel_id, (x, y) in enumerate(_iter_channel(positions)):
        masked = is_masked.get(channel_id, True)
        if masked:
            yield _disk(x, y, size_masked, color_masked)
            continue
        for clu in clusters_per_channel[channel_id]:
            # Translation.
            t = 10 * (clu - .5 * (n_clusters - 1))
            yield _disk(x, y, size_unmasked, color_umasked(clu), t=t)


def probe_layout(positions, cluster_channels):
    """Return a probe layout in SVG, showing selected clusters on relevant channels."""
    contents = '\n'.join(_iter_disks(positions, cluster_channels))
    return """
    <svg style="background: black; width:100%; height:100%;">
      {}
    </svg>
    """.format(contents)
